<?hh // strict

/**
 *
 * DB-IP.com database query and management class
 *
 * Copyright (C) 2016 db-ip.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

class DBIP_Exception extends Exception {

}

class DBIP_Record {
	public ?string $ip_start;
	public ?string $ip_end;
	public ?string $country;
	public ?string $stateprov;
	public ?string $city;
	public ?float $latitude;
	public ?float $longitude;
	public ?float $timezone_offset;
	public ?string $timezone_name;
	public ?string $isp_name;
	public ?string $connection_type;
	public ?string $organization_name;
}

class DBIP {

	const int TYPE_COUNTRY = 1;
	const int TYPE_CITY = 2;
	const int TYPE_LOCATION = 3;
	const int TYPE_ISP = 4;
	const int TYPE_FULL = 5;
	
	private PDO $db;
	
	public function __construct(PDO $db) {
		$this->db = $db;
		$this->db->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_EXCEPTION);
	}

	public function Import_From_CSV(string $filename, string $type, string $table_name = "dbip_lookup", ?(function(int): void) $progress_callback = null): int {
		switch ($type) {
			case self::TYPE_COUNTRY:
			$fields = [ "ip_start", "ip_end", "country" ];
			break;
			case self::TYPE_CITY:
			$fields = [ "ip_start", "ip_end", "country", "stateprov", "city" ];
			break;
			case self::TYPE_LOCATION:
			$fields = [ "ip_start", "ip_end", "country", "stateprov", "district", "city", "zipcode", "latitude", "longitude", "geoname_id", "timezone_offset", "timezone_name" ];
			break;
			case self::TYPE_ISP:
			$fields = [ "ip_start", "ip_end", "country", "isp_name", "connection_type", "organization_name" ];
			break;
			case self::TYPE_FULL:
			$fields = [ "ip_start", "ip_end", "country", "stateprov", "district", "city", "zipcode", "latitude", "longitude", "geoname_id", "timezone_offset", "timezone_name", "isp_name", "connection_type", "organization_name" ];
			break;
			default:
			throw new DBIP_Exception("invalid database type");
		}
		if (!file_exists($filename)) {
			throw new DBIP_Exception("file {$filename} does not exists");
		}
		$q = $this->Prepare_Import($table_name, $fields);
		if (substr($filename, -3) === ".gz") {
			$openname = "compress.zlib://" . $filename;
		} else {
			$openname = $filename;
		}
		$f = @fopen($openname, "r");
		if (!is_resource($f)) {
			throw new DBIP_Exception("cannot open {$filename} for reading");
		}
		$nrecs = 0;
		while ($r = fgetcsv($f, 1024, ",", '"'/*, "\0"*/)) {
			foreach ($r as $k => $v) {
				if ($fields[$k] === null) {
					throw new DBIP_Exception("invalid CSV record for {$r[0]}");
				}
				switch ($fields[$k]) {
					case "connection_type":
					$r[$k] = $v ?: null;
					break;
					case "latitude":
					case "longitude":
					case "timezone_offset":
					$r[$k] = (float)$v;
					break;
					case "isp_name":
					case "organization_name":
					$r[$k] = substr($v, 0, 128);
					break;
					case "city":
					case "district":
					case "stateprov":
					$r[$k] = substr($v, 0, 80);
					break;
					case "zipcode":
					$r[$k] = substr($v, 0, 20);
					break;
					case "timezone_name":
					$r[$k] = substr($v, 0, 64);
					break;
					case "geoname_id":
					if (!$r[$k]) $r[$k] = null;
					break;
					default:
					$r[$k] = stripslashes($v);
				}
			}
			$r[] = self::Addr_Type($r[0]);
			$r[0] = inet_pton($r[0]);
			$r[1] = inet_pton($r[1]);
			$this->Import_Row($q, $r);
			if ((($nrecs % 100) === 0) && ($progress_callback !== null)) {
				call_user_func($progress_callback, $nrecs);
			}
			$nrecs++;
		}
		fclose($f);
		$this->Finalize_Import();
		return $nrecs;
	}

	public function Lookup(string $addr, string $table_name = "dbip_lookup"): DBIP_Record {
		if ($ret = $this->Do_Lookup($table_name, self::Addr_Type($addr), inet_pton($addr))) {
			$ret->ip_start = inet_ntop($ret->ip_start);
			$ret->ip_end = inet_ntop($ret->ip_end);
			return $ret;
		} else {
			throw new DBIP_Exception("address not found");
		}
	}

	protected function Prepare_Import(string $table_name, array<string> $fields): PDOStatement {
		try {
			if ($this->db->query("select count(*) as cnt from `{$table_name}`")->fetchObject()->cnt) {
				throw new DBIP_Exception("table {$table_name} is not empty");
			}
			$this->db->setAttribute(PDO::ATTR_EMULATE_PREPARES, false);
			$this->db->beginTransaction();
			$q = $this->db->prepare("insert into `{$table_name}` (" . implode(",", $fields) . ",addr_type) values (" . implode(",", array_fill(0, count($fields), "?")) . ",?)");
			return $q;
		} catch (PDOException $e) {
			throw new DBIP_Exception("database error: {$e->getMessage()}");
		}
	}

	protected function Finalize_Import(): void {
		$this->db->commit();
	}

	protected function Import_Row(PDOStatement $res, array<mixed> $row): bool {
		return $res->execute($row);
	}

	protected function Do_Lookup(string $table_name, string $addr_type, string $addr_start): DBIP_Record {
		$q = $this->db->prepare("select * from `{$table_name}` where addr_type = ? and ip_start <= ? order by ip_start desc limit 1");
		$q->execute(array($addr_type, $addr_start));
		return $q->fetchObject(DBIP_Record::class);
	}
	
	static private function Addr_Type(string $addr): string {
		if (ip2long($addr) !== false) {
			return "ipv4";
		} else if (preg_match('/^[0-9a-fA-F:]+$/', $addr) && @inet_pton($addr)) {
			return "ipv6";
		}
		throw new DBIP_Exception("unknown address type for {$addr}");
	}

}